;; Attempt to create a simple text adventure.
;; Marijn Haverbeke
;; Started June 8th 2004

;; This file contains all of the game code for the Dunwich text
;; adventure game. Some specific code has been put in separate files
;; (see below) to keep the size of this one reasonable. A lot of the
;; code in this file is very repetitive, it specifies the game world
;; and specific interactions with it. It is also really ugly because
;; of all the long strings in it. I probably should have made a
;; special string table, but I did not. If your editor is less than
;; 230 characters wide, get ready for some really ugly line-breaking.

;; The code uses as little nonstandard scheme code as possible, but on
;; a few points I really had to use mzscheme functions. Firstly
;; (flush-output) is used to flush the output stream after the prompt
;; has been displayed, otherwise the prompt would appear after the
;; user enters text. Secondly, I use the (random <n>) to generate
;; random numbers. Thirdly, the load and save functions in
;; load-save.scm use some nonstandard file manipulation functions. The
;; standard support for files and streams in scheme is pathetic.

;; INCLUDES

; Make errors informative.
; (require (lib "errortrace.ss" "errortrace"))

; In- and output functions (read-input-line and print-message).
(load "in-out.scm")
; The table system for our data.
(load "nested-tables.scm")
; The command parser.
(load "parse.scm")
; Loading and saving.
(load "load-save.scm")

;; MAIN GAME LOOP

; Main game function. Drops you into the game. If you are in an
; interactive scheme process you can call (game) again after leaving
; the game to return to the point where you were before.
(define (game)
  (welcome-message)
  (move-to (current-location))
  (main-loop))

; The intro story.
(define (welcome-message)
  (if (get-state 'run-before)
      (print-message "Welcome back.")
      (begin (set-state 'run-before #t)
	     (print-message "Welcome to the South End of the city of Dunwich. Your long search for eternal life has "
			    "currently brought you to this derelict neighborhood. A week ago some dodgy character told "
			    "you there was a group of people living around here who possessed the elixir of life, the secret "
			    "of eternal life. Why such a group would choose such an awful place to spend their eternal lives "
			    "is unclear to you. Anyway, try to find that elixir. Type 'help' to get some help."))))

; This function takes care of the main loop. It shows a prompt and
; reads input until interpret returns false with the input.
(define (main-loop)
  (define (get-proper-input)
    (let ((input (read-input-line)))
      (if (null? input)
	  (get-proper-input)
	  input)))
  (run-turn-events)
  (display "# ")
  (flush-output)
  (if (interpret (get-proper-input))
      (main-loop)))

; Try to parse the input. When successful, call the returned function
; on the returned arguments.
(define (interpret input)
  (let ((parsed (parse input (action-patterns))))
    (if parsed
	(call-with (car parsed) (cadr parsed))
	(print-message
	 (list-ref '("Sorry, I do not understand you..."
		     "Use simpler sentences, I am only a computer."
		     "I can not make sense of that."
		     "Try to word that differently."
		     "I could not recognize that sentence.")
		   (random 5))))))

;; ACTIONS

;; Functions that are called as a response to user input.

(define (look-action . args)
  (if (null? args)
      (display-object (find-object (current-location)))
      (let ((found (find-object (car args))))
	(cond ((not found) (print-message "error: unknown object"))
	      ((object-is-here (car args)) (display-object found))
	      (else (print-message (capitalize (object-string found)) " is not here."))))))

(define (display-object object)
  (define (enumerate-objects objects)
    (let ((words (map find-name objects)))
      (cond ((null? objects) #t)
	    ((null? (cdr objects)) (print-message "You see " (car words) " here."))
	    (else (let ((message (append (enumerate words) '(" are here."))))
		    (set-car! message (capitalize (car message)))
		    (apply print-message message))))))
  (cond ((eq? (find-object (current-location)) object)
	 (execute-action object 'look-loc (lambda () (print-message "You are at " (object-string object) ".")))
	 (enumerate-objects (filter (lambda (x) (member (object-type (find-object x)) '(npc movable))) (objects-at object))))
	(else (execute-action object 'look (lambda () (print-message "You see nothing special about " (object-string object) "."))))))

(define (look-close-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((object-is-here object) (execute-action found 'look-close (lambda () (print-message "You see nothing of interest."))))
	  (else (print-message (capitalize (object-string found)) "is not here.")))))

(define (take-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((member object (inventory)) (print-message "You already have " (object-string found)))
	  ((not (object-is-here object)) (print-message (capitalize (object-string found)) " is not here."))
	  ((member (object-type found) '(location npc)) (print-message "You can not take that!"))
	  ((eq? (object-type found) 'prop) (execute-action found 'get (lambda () (print-message "Nope, too heavy."))))
	  (else (take-object-from object (current-location))))))

(define (take-object-from object origin)
  (add-to-inventory object)
  (remove-object-from (find-object origin) object)
  (print-message "Taken."))

(define (drop-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((not (member object (inventory))) (print-message "You do not have " (object-string found)))
	  (else (drop-object-at object (current-location)) (print-message "Done.")))))

(define (drop-object-at object place)
  (remove-from-inventory object)
  (put-object-at (find-object place) object))

(define (say-action sentence)
  (cond ((equal? sentence '(xyzzy)) (print-message "You utter a magic word. Magic was outlawed centuries ago."))
	(else (print-message "You talk, and nobody listens (Try saying things TO people)."))))

(define (talk-action sentence target)
  (let ((found (find-object target)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((not (object-is-here target))
	   (print-message "You do not see " (object-string found) " anywhere."))
	  ((not (eqv? (object-type found) 'npc))
	   (print-message "Please stop talking to inanimate objects. People are staring at you."))
	  (else 
	   (execute-action found 'talk (lambda () (print-message (capitalize (object-string found)) " does not respond.")) sentence)))))

(define (climb-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((not (object-is-here object)) (print-message "You'd have to be near " (object-string found) " to climb it."))
	  (else (execute-action found 'climb (lambda () (print-message "It is not possible to climb " (object-string found) ".")))))))

(define (open-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((not (object-is-here object)) (print-message (capitalize (object-string found)) " is not here."))
	  (else (execute-action found 'open (lambda () (print-message "It is not possible to open " (object-string found) ".")))))))

(define (put-in-action object target)
  (let ((found-object (find-object object)) (found-target (find-object target)))
    (define (impossible) (print-message "That does not work."))
    (cond ((not (and found-object found-target)) (print-message "error: unknown object"))
	  ((not (object-is-here object)) (print-message "You do not have " (object-string found-object) "."))
	  ((not (object-is-here target)) (print-message (capitalize (object-string found-target)) " is not here."))
	  ((execute-action found-target 'put-in impossible object))
	  (else (impossible)))))

(define (give-action object target)
  (let ((found-object (find-object object)) (found-target (find-object target)))
    (define (refusal) 
      (print-message (capitalize (object-string found-target)) " politely refuses " (object-string found-object) "."))
    (cond ((not (and found-object found-target)) (print-message "error: unknown object"))
	  ((not (member object (inventory))) (print-message "You do not have " (object-string found-object) "."))
	  ((not (eqv? (object-type found-target) 'npc)) (print-message "You can only give things to people."))
	  ((not (object-is-here target)) (print-message (capitalize (object-string found-target)) " is not here."))
	  ((execute-action found-target 'accept refusal object))
	  (else (refusal)))))

(define (load-action file)
  (if (load-game (car file))
      (look-action)))

(define (save-action file)
  (save-game (car file)))

(define (inventory-action)
  (let ((inventory (map find-name (inventory))))
    (cond ((null? inventory) (print-message "You do not own anything."))
	  ((null? (cdr inventory)) (print-message "You've got " (car inventory) "."))
	  (else (apply print-message (append '("You are carrying ") (enumerate inventory) '(".")))))))

(define (help-action)
  (print-message "This game follows most of the conventions of text adventures and interactive fiction. you perform actions "
		 "by typing in simple imperative sentences like 'look at the ferry' or 'climb fence', most of which can be "
		 "shortened to things like 'l ferry'. To see what you are currently carrying type 'inventory' or 'i', to leave "
		 "the game type 'exit'. Moving around goes a little different in this game than in most similar games. Instead "
		 "of choosing a direction you say things like 'enter warehouse' or 'go to avenue' to move somewhere. You can "
		 "go to locations that are not adjacent to your current location this way (if you have discovered them).")
  (print-message "Conversation with NPCs goes as follows: NPCs are rather dumb, all they really do is respond to 'magic words'. "
		 "A good way to start a conversation is to begin with a greeting ('say \"hello\" to ferryman') or by asking "
		 "about the elixir of life ('ask ferryman about elixir'), then you can follow through the conversation by "
		 "asking about specific words the character used that sounded interesting.")
  (print-message "To save your game type 'save <gamename>', to load it again use 'load <gamename>'. The game will be saved as "
		 "<gamename>.save in the current directory."))

(define (read-action object)
  (let ((found (find-object object)))
    (cond ((not found) (print-message "error: unknown object"))
	  ((not (object-is-here object)) (print-message "You do not see " (object-string found) "."))
	  (else (execute-action found 'read (lambda () (print-message "There is nothing to read on that.")))))))

(define (leave-action)
  (let ((adjacent-locations (filter (lambda (x) (eq? (object-type x) 'location))
				    (map find-object (objects-at (find-object (current-location)))))))
    (if (null? adjacent-locations)
	(print-message "There is no way out!")
	(move-action (object-id (car adjacent-locations))))))
    

(define (move-action destination)
  (let ((found (find-object destination)))
    (cond ((or (not found)
	       (not (eq? (object-type found) 'location))
	       (not (location-is-known found))) (print-message "You know no such location."))
	  (else (move-to destination)))))

(define (move-to location)
  (cond ((execute-action (find-object location) 'enter do-nothing)
	 (set-current-location location)
	 (look-action))))  

(define (quit-action)
  (print-message "Bye!")
  #f)

(define (execute-action object action default . args)
  (let ((found (object-action object action)))
    (if found
	(call-with found args)
	(default))))

;; MISC GAME FUNCTIONS

;; Some game details.

(define (words-in-sentence? words sentence)
  (define (loop sentence)
    (cond ((null? sentence) #f)
	  ((member (car sentence) words) #t)
	  (else (loop (cdr sentence)))))
  (loop sentence))

(define (talk-to-worker sentence)
  (print-message "The worker looks up and says 'sorry, can't talk, working.'")
  (cond ((not (get-state 'overseer-entered))
	 (put-object-at (find-object 'cloth-factory) 'overseer)
	 (print-message "Hearing you talk to one of the workers, the overseer comes from behind a corner to see what is "
			"happening. He asks you 'What are you doing here?'")
	 (set-state 'overseer-entered #t))))

(define (talk-to-ferryman sentence)
  (cond ((words-in-sentence? '(elixer elixir life potion alchemy alchemist) sentence) 
	 (print-message "The ferryman says 'Elixir? Alchemy? I don't know, you should ask the overseer at the factory.'"))
	((words-in-sentence? '(factory) sentence) 
	 (print-message "'The cloth factory, it's over at Gray Square.'"))
	((words-in-sentence? '(overseer) sentence) 
	 (print-message "The ferryman responds 'The overseer in the cloth factory. He's a weird guy, I think I heard he does "
			"alchemy.'"))
	((words-in-sentence? '(ferry boat take Dunwich town) sentence) 
	 (print-message "The ferryman looks at you and frowns. 'I'm not taking anyone back until the end of the day when the "
			"factory closes.'"))
	((words-in-sentence? '(hello hi) sentence) 
	 (print-message "The ferryman returns your greeting, but does not look any more friendly."))
	((words-in-sentence? '(pyromaniac fire fires newspaper limits arson investigation) sentence) 
	 (print-message "The ferryman says 'There's been some trouble with a pyromaniac it seems. Funny stuff really. Ask the "
			"cop if you want to know more.'"))
	((words-in-sentence? '(funny) sentence) 
	 (print-message "'I don't know, some wacko running around torching stuff, funny to me. Ask the police guy if you want "
			"to know more.'"))
	((words-in-sentence? '(police cop) sentence)
	 (print-message "The ferryman sneers and says 'Yep, we've got our own police officer in South End now. You can't miss "
			"him, he patrols around all day.'"))
	((words-in-sentence? '(Harris) sentence) (print-message "The ferryman says 'Harris? He does pass by here now and then.'"))
	((words-in-sentence? '(island isle) sentence) (try-to-go-to-island))
	((and (words-in-sentence? '(dock docks south end) sentence) (eqv? (current-location) 'island)) (move-to 'riverside))
	(else (print-message "The ferryman responds 'I know nothing about that.'"))))

(define (try-to-go-to-island)
  (cond ((not (get-state 'pyromaniac-caught))
	 (print-message "The ferryman sighs. 'For the duration of the arson investigation the island is off limits. "
			"I'd get in trouble if I took anyone there.'"))
	((eqv? (current-location) 'island)
	 (print-message "The ferryman looks at you 'Eh? We are already on the island.'"))
	(else
	 (print-message "The ferryman responds 'All right, now that the pyromaniac is caught, I can take you there. "
			"Get in the boat.' He takes you to the island. (You can now move to and from the island as "
			"with other locations)")
	 (set-known 'island)
	 (move-to 'island))))

(define (try-to-open-door)
  (cond ((get-state 'door-is-open) (print-message "The door is already open."))
	(else (print-message "You push the door but it does not move."))))

(define (try-to-put-in-door object)
  (cond ((eqv? object 'needle)
	 (print-message "With a light 'click' the door opens.")
	 (set-state 'door-is-open #t))
	(else #f)))

(define (try-to-enter-barn)
  (cond ((not (get-state 'door-is-open))
	 (print-message "You'd have to get the door open first.")
	 #f)
	(else #t)))

(define (try-to-enter-factory)
  (cond ((not (get-state 'bribed-guard))
	 (print-message "The guard stops you. 'Listen, they pay me to keep strangers out of here. They do not pay me very much, "
			"but unless you are willing to increase my pay a little I'm not letting you in there.'")
	 #f)
	(else #t)))

(define (talk-to-guard sentence)
  (cond ((words-in-sentence? '(increase pay) sentence) 
	 (print-message "The guard says 'That's right, give me some money or something valuable and I'll look the other way'"))
	((words-in-sentence? '(watch pocketwatch) sentence)
	 (print-message "The guard looks at the watch for a moment. 'It's broken, but it'll do. Give it to me.'"))
	((words-in-sentence? '(enter factory through) sentence)
	 (print-message "The guard looks at you threateningly. 'You're not getting in. not for free at least.'"))
	((words-in-sentence? '(hi hello) sentence) (print-message "The guard squints. 'Hi.'"))
	(else (print-message "The guard ignores your babbling."))))

(define (guard-accept object)
  (cond ((eqv? object 'watch)
	 (remove-from-inventory 'watch)
	 (set-state 'bribed-guard #t)
	 (print-message "The guard grunts something that could be interpreted as 'thankyou' and pockets the watch."))
	(else #f)))

(define (talk-to-overseer sentence)
  (cond ((words-in-sentence? '(alchemy elixer elixir medicine stone) sentence)
	 (print-message "The overseer seems disturbed. 'I used to be somewhat, just a little mind you, involved with that "
			"alchemy stuff. It is all finished now though. Only Harris was so deep in it he never could quit.'"))
	((words-in-sentence? '(harris) sentence) 
	 (print-message "The overseer sighs. 'We used to have a little group here, an alchemist circle. It was childish really. "
			"Harris was the leader, kind of. He took it all very seriously. I have no idea where he is now, he "
			"hardly shows up in this neighborhood anymore. You could try asking the baker, he was a member longer "
			"than I was."))
	((words-in-sentence? '(hi hello) sentence) 
	 (print-message "'And a good day to you. Now leave.'"))
	((words-in-sentence? '(looking around) sentence)
	 (print-message "The overseer responds 'Man, I thought I told that no-good guard of ours to keep out useless louts "
			"like you.'"))
	((words-in-sentence? '(diary code symbols) sentence) 
	 (print-message "The overseer frowns. 'Must be the code those alchemists used in their notes. They never explained "
			"it to me.'"))
	(else (print-message "The overseer says 'I do not care. Please stop distracting the workers.'"))))

(define (talk-to-baker sentence)
  (cond ((words-in-sentence? '(bread buy) sentence) 
	 (print-message "The baker responds 'For 2 schilling you can have some.' You don't feel very hungry though."))
	((words-in-sentence? '(alchemy elixer elixir stone medicine) sentence) 
	 (print-message "The baker smiles wryly and says 'I did once dabble in alchemy. Kind of lost faith in it though.'"))
	((words-in-sentence? '(harris) sentence)
	 (print-message "'Yeah I know Harris. He used to be a major guy around here. Now he rarely shows up. Last time I saw "
			"him he was coming out of the lumber mill, god knows what he was doing there.'"))
	((words-in-sentence? '(hello hi) sentence) 
	 (print-message "The baker smiles and says 'Good day to you.'"))
	((words-in-sentence? '(faith) sentence) 
	 (print-message "The baker replies 'Well, now that all the scientists are saying that the whole lead to gold thing is "
			"impossible, many alchemists are getting pessimistic.'"))
	((words-in-sentence? '(pyromaniac fire fires newspaper) sentence)
	 (print-message "The baker replies 'There have been a lot of fires in Dunwich lately. people say it's a pyromaniac.'"))
	((words-in-sentence? '(diary) sentence)
	 (print-message "The baker responds 'That must be Harris' diary. We used to have a special code to write down our notes. "
			"Mostly to keep up an appearance of mysteriousness.'"))
	((words-in-sentence? '(code symbols) sentence)
	 (print-message "The baker thinks for a moment and then says 'Sure I can teach you the code. No one really cares about "
			"it anymore anyway.' He explains the system to you, which turns out to be really simple.")
	 (set-state 'know-diary-code #t))
	(else (print-message "The baker looks puzzled 'Sorry, I can't help you with that.'"))))

(define (talk-to-bum sentence)
  (cond ((words-in-sentence? '(hi hello) sentence)
	 (print-message "The bum grins at you thankfully and replies 'Good day. Allow me to introduce myself. I am Jean-Paul "
			"van Ricken, former owner of the Dunwich Lumber Company.'"))
	((words-in-sentence? '(bandage finger) sentence)
	 (print-message "The bum looks at the bandage and responds 'Ehh well, I cut my finger on some glass yesterday. Nothing "
			"serious.'"))
	((words-in-sentence? '(fire fires pyromaniac newspaper) sentence)
	 (print-message "A gleam appears in the eyes of the bum. 'Someone's been starting fires. But who?'"))
	((words-in-sentence? '(ricken) sentence) 
	 (print-message "'My name is Jean-Paul van Ricken. I come from a rich and noble family.'"))
	((words-in-sentence? '(former owner lumber company) sentence) 
	 (print-message "The bum looks very sad now. 'I really did own the Lumber Company. But a year ago, my debts were "
			"so high they took the whole thing away from me. They say I ruined the company because I drank too "
			"much, but that is nonsense. It was a case sabotage and dirty play.' The bum stares at the ground and "
			"broods."))
	((words-in-sentence? '(sabotage dirty) sentence) 
	 (print-message "The bum responds 'True. My enemies ruined me because they were jealous.'"))
	((words-in-sentence? '(enemies jealous) sentence)
	 (print-message "The bum frowns and replies 'When you are successful, there are always those who can not stand it.'"))
	((words-in-sentence? '(wine booze alcohol drunk) sentence)
	 (print-message "The bum says 'I know, I drink too much. What would you do though, if you were ruined like I am?"))
	((words-in-sentence? '(elixer elixir alchemy alchemist alchemists) sentence)
	 (print-message "The bum looks uninteresed. 'I do not concern myself with alchemy.'"))
	((words-in-sentence? '(glove gloves can matches) sentence)
	 (print-message "The bum looks hostile now. 'Never seen 'em before. Bye.'"))
	((words-in-sentence? '(family) sentence)
	 (print-message "The bum sighs 'My family used to be rich. Now they are poor. They never want to see me again they say.'"))
	(else (print-message "The bum looks annoyed. 'I don't know about that.' Then he lightens up 'Hey, did you know I used "
			     "to own the Dunwich Lumber Company?'"))))

(define (talk-to-harris sentence)
  (cond ((words-in-sentence? '(hi hello) sentence)
	 (print-message "Harris replies 'Hello stranger, what brings you here?'"))
	((words-in-sentence? '(elixer elixir medicine) sentence)
	 (print-message "Harris looks at you closely. 'So you want to find the elixir of life? Well, I've got this book here "
			"that might help you.' He shows you a large book. 'but keep in mind, the search for the elixir is one "
			"for wisdom and purity, not for gold or eternal life.'"))
	((words-in-sentence? '(book) sentence)
	 (print-message "'Here. This book will teach you about the philosopher's stone.' He hands you the book.")
	 (add-to-inventory 'book))
	((words-in-sentence? '(alchemy alchemist alchemists) sentence)
	 (print-message "Harris smiles. 'Yes, I am a practicer of the Art. The only remaining alchemist in town.'"))
	((words-in-sentence? '(stone) sentence)
	 (print-message "Harris says 'You see, the philosopher's stone and the elixir of life are really the same thing, "
			"the stone is not really a stone, it is the essence of purity, the goal of alchemy.'"))
	((words-in-sentence? '(diary) sentence)
	 (print-message "'That must be my diary. I do not care much about such mundane things anymore these days.'"))
	((words-in-sentence? '(rose rosa cross croix) sentence)
	 (print-message "Harris responds 'The order of the rosa croix used to be a group of magnificent alchemists. "
			"I consider myself something of an apprentice of them.'"))
	(else
	 (print-message "Harris looks at you with deep understanding, but does not respond."))))

(define (read-the-book)
  (print-message "You sit down and start studying Harris' book. It is a rather recent translation of an ancient Arabic text. You "
		 "encounter various terms that seem to refer to the same thing, 'philosopher's stone', 'elixir', 'Medicine' and "
		 "'essence of life'. This essence is supposed to be something, you are unsure whether it is an actual substance or "
		 "just a state of mind, that brings out the perfection in anything. It turns lead to gold, confusion to "
		 "enlightenment, sickness to health, etc.")
  (print-message "The book goes on to describe the process required for obtaining the stone. In between an enormous amount of "
		 "obscure talk, of which you are not sure whether it discusses chemistry, rituals or philosophy, you think you "
		 "find the answer to your question. To obtain the stone, the ultimate wisdom, you have to combine two spheres of "
		 "thinking. One sphere is that of the sun, which refers to rational thinking, power and masculinity. On the other "
		 "hand is the sphere of the moon, the sphere of emotional thinking, melancholy and femininity. The sun sphere has "
		 "a tendency to grow arrogant and blind, while the moon sphere is by itself destined to be passive and suffer. "
		 "Only by combining the two can wisdom and understanding be achieved.")
  (print-message "somewhat enlightened, you go home.")
  #f)

(define (talk-to-officer sentence)
  (cond ((words-in-sentence? '(hello hi mind question fires fire pyromaniac newspaper) sentence)
	 (print-message "The officer says 'What I want to ask you is related to the fires we've been having in Dunwich "
			"lately. In the last ten months eight buildings were burned down, all were property of a small group "
			"of industrials. We have good reasons to believe the fires were started intentionally, most of them "
			"using lamp oil. We also believe that the perpetrator lives in South End. If you could keep your eyes "
			"open and report anything suspicious to me I would be very grateful.'"))
	((words-in-sentence? '(can lamp oil) sentence)
	 (if (object-is-here 'can)
	     (print-message "'An empty can of lamp oil? That has to have something to do with the fires. Please give me the "
			    "evidence.'")
	     (print-message "The officer says 'The fires were started with lamp oil.'")))
	((words-in-sentence? '(matches match) sentence)
	 (if (object-is-here 'matches)
	     (print-message "The officer looks at the matches. 'Some wet old matches, what is that supposed to prove?'")
	     (print-message "'Matches? What about them?'")))
	((words-in-sentence? '(gloves) sentence)
	 (if (object-is-here 'gloves)
	     (print-message "The officer inspects the gloves. 'Interesting. I consider this to be evidence. Can you give "
			    "it to me?'")
	     (print-message "'Gloves? Can you show them to me?'")))
	((words-in-sentence? '(bum jean paul ricken bandage) sentence)
	 (if (get-state 'proven-guilty)
	     (finish-pyromaniac-part)
	     (print-message "'So you suspect Jean-Paul? It is possible, but you will need some more evidence.'")))
	((words-in-sentence? '(alchemy alchemist elixer elixir) sentence)
	 (print-message "The officer thinks for a moment and says 'There used to be some people here who considered themselves "
			"alchemists. I haven't heard much of it lately, but I think the overseer at the factory could tell "
			"you more.'"))
	((words-in-sentence? '(harris) sentence)
	 (print-message "'Harris? Yes, I know him. A real intelligent and serious fellow. He does not come here very often "
			"anymore.'"))
	(else (print-message "The officer responds 'I have no idea what you are talking about.'"))))

(define (finish-pyromaniac-part)
  (set-state 'pyromaniac-caught #t)
  (remove-object-from (find-object 'trail) 'bum)
  (remove-object-from (find-object (current-location)) 'officer)
  (print-message "The officer nods. 'That makes sense, mr. Van Ricken's finger is indeed burned. I've always suspected "
		 "that he isn't entirely right in his mind. I'll go get him, consider the investigation closed. My thanks "
		 "for your help.'"))

(define (officer-accept object)
  (cond ((eqv? object 'gloves)
	 (print-message "The officer takes the gloves and says 'Thank you, this will be useful. Did you notice the little "
			"finger of the left glove is burnt?'")
	 (remove-from-inventory 'gloves)
	 (set-state 'given-gloves #t)
	 (if (get-state 'given-can) (set-state 'proven-guilty #t)))
	((eqv? object 'can)
	 (print-message "The officer accepts the can. 'Thanks. Look here, there's a bramble leaf sticking to it.'")
	 (remove-from-inventory 'can)
	 (set-state 'given-can #t)
	 (if (get-state 'given-gloves) (set-state 'proven-guilty #t)))
	(else #f)))

(define (move-officer)
  (cond ((and (not (object-is-here 'officer)) (= (random 3) 1))
	 (remove-object-from (find-object 'industry-avenue) 'officer)
	 (remove-object-from (find-object 'gray-square) 'officer)
	 (remove-object-from (find-object 'riverside) 'officer)
	 (let ((new-location (list-ref '(industry-avenue gray-square riverside) (random 3))))
	   (put-object-at (find-object new-location) 'officer)
	   (if (eqv? new-location (current-location))
	       (print-message "The police officer walks into view."))))))

(define (try-to-climb-attic)
  (if (try-to-enter-attic)
      (move-to 'attic)))

(define (try-to-enter-attic)
  (cond ((member 'ladder (inventory))
	 (drop-object-at 'ladder 'lumber-mill)
	 (print-message "Using the ladder you can easily enter the attic.")
	 #t)
	((member 'ladder (objects-at (find-object 'lumber-mill)))
	 #t)
	(else
	 (print-message "You can not reach it. You'd need something to stand on.")
	 #f)))

(define (try-to-read-diary)
  (cond ((not (object-is-here 'diary)) (print-message "You do not have the diary"))
	((get-state 'know-diary-code)
	 (print-message "Now that you understand the code you can decipher the diary. It seems that it is indeed the diary of "
			"Harris. it contains six entries. Which one do you want to read? (Hint: type 'read first entry')"))
	(else (print-message "You do not understand these symbols. Try to find someone who does."))))

(define (read-diary-entry number)
  (cond ((not (object-is-here 'diary)) (print-message "You do not have the diary."))
	((not (get-state 'know-diary-code)) (print-message "You can not read the diary."))
	((member (car number) '(one first))
	 (print-message "This is the first entry in the diary. It was written two years and nine months ago, only a short time "
			"after Harris had decided to study alchemy. He stresses that he is more interested in cleansing his "
			"soul than in cleansing metal. You assume that means transmuting metal to gold."))
	((member (car number) '(two second))
	 (print-message "This is the second entry in the diary. Written two years and six months ago. Harris reports how he has "
			"formed a little group of alchemists and enthusiastically describes their meetings and discussions. The "
			"baker and a few people you have never heard of are mentioned. Still nothing that is very relevant."))
	((member (car number) '(three third))
	 (print-message "This is the third entry in the diary. Written two years ago. It tells the story of a big fight between "
			"two parts of the alchemists circle. Apparently the 'senior' members came into conflict with newer "
			"member about priorities in alchemy. The new people are mostly interested in making gold. the seniors "
			"abhor this. It appears that the overseer from the factory belonged to this group of new members."))
	((member (car number) '(four fourth))
	 (print-message "This is the fourth entry in the diary. It was written one year and eight months ago. There is no more "
			"mention of the conflict in this entry. It mostly deals with Harris' search for something called the "
			"'philosopher's stone', describing various attempts, both chemical and spiritual, to 'achieve' it."))
	((member (car number) '(five fifth))
	 (print-message "This is the fifth entry in the diary. Written one year and five months ago. The alchemists are very "
			"upset with developments in science that seem to indicate that transmuting metal to gold is impossible. "
			"The opinion that most historical alchemists were charlatans is taking possession of the circle and the "
			"group is crumbling."))
	((member (car number) '(6 six sixth))
	 (print-message "This is the sixth entry in the diary. Written one year and two months ago. The alchemists circle has "
			"dissolved. Harris seems not to mind that too much. He decides to continue his search for the "
			"philosopher's stone alone and in seclusion. He seems very happy about a new 'laboratory' he found, "
			"apparently it is situated on an island. He mentions how it's door will open 'only to the tiniest key'. "
			"This is the last entry, there is no clue as to what Harris might have been doing this last year. "
			"While it seems unlikely that he is currently in the possession of the elixir, you'd like to find him "
			"to make sure."))
	(else (print-message "There is no entry '" (car number) ".'"))))

(define (look-close-at-desk)
  (cond ((get-state 'found-can)
	 (print-message "There is nothing else to be seen behind the desk."))
	(else
	 (print-message "Behind the desk, in a little alcove in the wall, lies an empty can.")
	 (put-object-at (find-object 'warehouse) 'can)
	 (set-state 'found-can #t))))

(define (look-close-at-brambles)
  (cond ((get-state 'found-gloves)
	 (print-message "There is nothing else under the bramble bushes."))
	(else
	 (print-message "Under the bramble bushes you find a set of matches and a pair of gloves.")
	 (put-object-at (find-object 'bridge) 'matches)
	 (put-object-at (find-object 'bridge) 'gloves)
	 (set-state 'found-gloves #t))))
  
(define (look-in-cabinet)
  (print-message "This cabinet looks like it once was sturdy and elegant. Now it is dusty and worm-eaten.")
  (cond ((not (get-state 'found-newspaper))
	 (print-message "You open it but the only thing inside is an old newspaper.")
	 (put-object-at (find-object 'warehouse) 'newspaper)
	 (set-state 'found-newspaper #t))))

;; ACTION FUNCTIONS

;; Standard functions to attach to objects to model their responses to
;; user actions.

(define (message-function . args)
  (lambda () (apply print-message args)))

(define (explore-function . args)
  (lambda () (for-each set-known args)))

(define (run-turn-events)
  (for-each (lambda (x) (x)) (get-turn-events)))

;; ENTITY MANIPULATION

;; Querying and modifying the entity data at a higher level. These
;; functions are the only ones that know about the actual
;; representation of the data. Currently it is one big nested table
;; (as implemented in nested-tables.scm) containing various branches
;; for the different kinds of data.

; Data is a global which forms the root to every other state variable
; in the program.
(define data (create-table))

(define (reset-data) (set! data (create-table)))

; A function to get the root of the data. Just in case I decide to
; store it somewhere else than in a global.
(define (get-data) data)

; Below here no more direct references to the data symbol.

; State variables are used for keeping track of things like whether
; events happened already or not.
(define (get-state symbol)
  (lookup-in-table (list 'state symbol) (get-data)))
(define (set-state symbol value)
  (insert-in-table (list 'state symbol) value (get-data)))

(define (define-object object-name name type)
  (insert-in-table (list 'object object-name 'name) name data)
  (insert-in-table (list 'object object-name 'type) type data)
  (lookup-in-table (list 'object object-name) (get-data)))
(define (find-object object-id)
  (lookup-in-table (list 'object object-id) (get-data)))

; The place where the player currently is.
(define (current-location) (cdr (lookup-in-table '(location) (get-data))))
(define (set-current-location location)
  (insert-in-table '(location) location (get-data))
  (set-known location))

; The stuff the player is carrying.
(define (inventory)
  (map car (filter (lambda (x) (cdr x)) (cdr (lookup-in-table '(inventory) (get-data))))))
(define (add-to-inventory object)
  (insert-in-table (list 'inventory object) #t (get-data)))
(define (remove-from-inventory object)
  (insert-in-table (list 'inventory object) #f (get-data)))

; Patterns to recognize commands.
(define (define-action-pattern pattern function)
  (insert-in-table (list 'action-patterns pattern) function (get-data)))
(define (action-patterns)
  (cdr (lookup-in-table '(action-patterns) (get-data))))

; Patterns for objects.
(define (define-object-pattern pattern symbol)
  (insert-in-table (list 'object-patterns pattern) symbol (get-data)))
(define (object-patterns)
  (cdr (lookup-in-table '(object-patterns) (get-data))))

; Query objects.
(define (object-type object)
  (cdr (lookup-in-table '(type) object)))
(define (object-string object)
  (cdr (lookup-in-table '(name) object)))
(define (object-id object)
  (car object))
(define (find-name object)
  (object-string (find-object object)))

; Actions (functions) associated with objects.
(define (add-action object action function)
  (cond ((lookup-in-table (list 'action action) object) => (lambda (x) (set-cdr! x (combine-functions (cdr x) function))))
	(else (insert-in-table (list 'action action) function object))))
(define (object-action object action)
  (let ((found (lookup-in-table (list 'action action) object)))
    (if found
	(cdr found)
	#f)))

; The location of objects.
(define (put-object-at location object)
  (insert-in-table (list 'present-objects object) #t location))
(define (remove-object-from location object)
  (insert-in-table (list 'present-objects object) #f location))
(define (object-is-here object)
  (cond ((or (eq? object (current-location)) (eq? object 'myself)) #t)
	((member object (inventory)) #t)
	((member object (objects-at (find-object (current-location)))) #t)
	(else #f)))
(define (objects-at location)
  (let ((found (lookup-in-table '(present-objects) location)))
    (if found
	(map car (filter (lambda (x) (cdr x)) (cdr found)))
	'())))

; Control whether locations are known to the player.
(define (set-known location)
  (insert-in-table '(known) #t (find-object location)))
(define (location-is-known location)
  (lookup-in-table '(known) location))

(define (add-turn-event function name)
  (insert-in-table (list 'turn-events name) function (get-data)))
(define (get-turn-events)
  (map cdr (cdr (lookup-in-table '(turn-events) (get-data)))))

;; MISC
;; Some helper functions.

(define (call-with function args)
  (if (null? args)
      (function)
      (apply function args)))

(define (do-nothing)
  (lambda () #t))

(define (combine-functions fun1 fun2) (lambda () (fun1) (fun2)))

(define (enumerate words)
  (if (null? (cddr words))
      (list (car words) " and " (cadr words))
      (append (list (car words) ", ") (enumerate (cdr words)))))

(define (filter predicate list)
  (cond ((null? list) '())
	((predicate (car list)) (cons (car list) (filter predicate (cdr list))))
	(else (filter predicate (cdr list)))))

(define (collector)
  (let ((collection '()))
    (lambda args
      (if (null? args)
	  collection
	  (set! collection (append collection args))))))

;; INITIALIZATION

;; Filling up the data structure. Looking back I would probably have
;; saved time by making some kind of meta-language for this, but the
;; way it is currently done (lots of functions that add elements and
;; properties to the world) is workable.

(define (init-world)
  ; A bunch of short forms for the often-used functions.
  (define (def-loc symbol name) (define-object symbol name 'location))
  (define (def-npc symbol name) (define-object symbol name 'npc))
  (define (def-mov symbol name) (define-object symbol name 'movable))
  (define (def-prop symbol name) (define-object symbol name 'prop))
  (define obj-pat define-object-pattern)
  (define act add-action)
  (define put put-object-at)
  (define msg message-function)

  (let ((myself (def-npc 'myself "yourself")))
    (obj-pat '(or me myself) 'myself)
    (act myself 'look (msg "You are looking shoddy as usual."))
    (act myself 'talk (lambda (x) (print-message "You have a long and pleasant conversation with yourself, but do not learn "
						 "anything new."))))

  (let ((watch (def-mov 'watch "the pocket watch")))
    (obj-pat '(seq (opt the) (opt pocket) (or watch pocketwatch clock)) 'watch)
    (act watch 'look (msg "This is a watch that you stole from your brother. It is broken but has great emotional value."))
    (add-to-inventory 'watch))

  (let ((crates (def-prop 'crates "the crates")))
    (obj-pat '(seq (opt the) (opt (seq pile of)) (or crate crates pile piles)) 'crates)
    (act crates 'look (msg "According to the labels, the crates do indeed contain woolen cloth.")))

  (let ((ships (def-prop 'ships "the ships")))
    (obj-pat '(seq (opt the) (or ship ships boat boats)) 'ships)
    (act ships 'look (msg "There three ships lying here. Two are meant for transporting cargo, and one is the ferry. The people "
			  "on the cargo ships do not look as if they would appreciate it if you were to go on board.")))

  (let ((river (def-prop 'river "the Urth river")))
    (obj-pat '(seq (opt the) (opt urth) (or river stream) (opt urth)) 'river)
    (act river 'look (msg "The Urth river is a rather short stream that transports water from the hills landinward to the North "
			  "Sea. Here, close to the sea, it is quite broad. You can see a small island a little downstream.")))

  (let ((trees (def-prop 'trees "the small trees")))
    (obj-pat '(seq (opt the) (opt small) (opt unhealthy) (or tree trees)) 'trees)
    (act trees 'look (msg "These trees look like they are about to die. Must be all the pollution in the river."))
    (act trees 'get (msg "Please leave the poor trees alone.")))

  (let ((sign (def-prop 'sign "the sign")))
    (obj-pat '(seq (opt the) sign) 'sign)
    (act sign 'look (msg "The sign says 'Laboratory. Keep out.'"))
    (act sign 'read (msg "'Laboratory. Keep out.'"))
    (act sign 'get (msg "It is nailed to the wall.")))

  (let ((door (def-prop 'door "the door")))
    (obj-pat '(seq (opt the) (opt (or barn laboratory)) (or door hole)) 'door)
    (act door 'look (msg "This is a sturdy wooden door with no doorknob. You can see a tiny little hole in the door on the place "
			 "where the doorknob should have been."))
    (act door 'open try-to-open-door)
    (act door 'put-in try-to-put-in-door))

  (let ((island (def-loc 'island "the island")))
    (obj-pat '(seq (opt the) island) 'island)
    (act island 'look (msg "A little island lying close to the other shore of the river. It has some small twisted trees on it."))
    (act island 'look-loc (msg "You are on the island in the Urth. Some unhealthy trees stand here, and on the back of the "
			       "island you notice a small barn."))
    (act island 'enter (explore-function 'barn))
    (put island 'trees)
    (put island 'barn)
    (put island 'sign)
    (put island 'door)
    (put island 'ferry)
    (put island 'ferryman)
    (put island 'river))

  (let ((harris (def-npc 'harris "Harris")))
    (obj-pat 'harris 'harris)
    (act harris 'look (msg "Harris is a white-haired man with spectacles. The oversized reddish shirt he is wearing gives him "
			   "a somewhat wizard-like appearance. He looks at you curiously."))
    (act harris 'talk talk-to-harris))

  (let ((book (def-mov 'book "the alchemy book")))
    (obj-pat '(seq (opt the) (opt large) (opt (or alchemy alchemist)) book) 'book)
    (act book 'look (msg "This is a heavy leather-bound book. On the front it says 'The Journey To The Stone Of The Philosopher'."))
    (act book 'read read-the-book))

  (let ((instruments (def-prop 'instruments "the instruments")))
    (obj-pat '(seq (opt the) (opt bizarre) (opt alchemy) (or instrument instruments)) 'instruments)
    (act instruments 'look (msg "Some of these instruments are recognizable as chemistry tools, but others make "
				"no sense whatsoever to you."))
    (act instruments 'get (msg "Harris does not appreciate it if you touch his alchemy equipment.")))

  (let ((pattern (def-prop 'pattern "the pattern on the floor")))
    (obj-pat '(seq (opt the) (opt strange) (or pattern floor)) 'pattern)
    (act pattern 'look (msg "The many little spots of light on the floor seem to form the shape of a flower, probably "
			    "a rose."))
    (act pattern 'get (msg "Not possible.")))

  (let ((barn (def-loc 'barn "the barn")))
    (obj-pat '(seq (opt the) (opt small) (or barn laboratory)) 'barn)
    (act barn 'look (msg "This is an old wooden barn, but someone has recently been working on it. A somewhat tower-like wooden "
			 "construction is standing on top of it and the door is heavily reinforced. Above the door hangs a sign."))
    (act barn 'look-loc (msg "You are inside an alchemist's laboratory. All kinds of bizarre instruments line the walls. Through "
			     "the windows in the tower some sunrays fall in, making a strange pattern on the rough wooden floor."))
    (act barn 'enter try-to-enter-barn)
    (put barn 'harris)
    (put barn 'instruments)
    (put barn 'island)
    (put barn 'pattern)
    (put barn 'door))

  (let ((ferry (def-prop 'ferry "the ferry")))
    (obj-pat '(seq (opt the) ferry) 'ferry)
    (act ferry 'look (msg "A rowing boat meant to transport people between South End and the rest of Dunwich. Ever since "
			  "they built the railroad the only connection between this neighborhood and the rest of the worlds "
			  "is over the water.")))

  (let ((ferryman (def-npc 'ferryman "the ferryman")))
    (obj-pat '(seq (opt the) (or (seq ferry man) ferryman)) 'ferryman)
    (act ferryman 'look (msg "This grizzled man has been rowing people back and forth for the last twenty years. He looks "
			     "strong as an ox and a lot grumpier than any ox you ever saw."))
    (act ferryman 'talk talk-to-ferryman))

  (let ((riverside (def-loc 'riverside "the riverside")))
    (obj-pat '(seq (opt the) (or riverside (seq river side) waterfront dock docks)) 'riverside)
    (act riverside 'look-loc (msg "You are at the riverside. Beside you the grey-brown stream of the Urth river slowly passes "
				  "the stone docks. Some boats are lying off the shore, calmly bobbing on the water. One of them "
				  "seems to be a ferry. Piles of small crates, probably containing cloth, line the docks. Poorly "
				  "built warehouses surround you. One of them is standing open. Industry Avenue leads landinward."))
    (act riverside 'enter (explore-function 'warehouse 'industry-avenue))
    (put riverside 'crates)
    (put riverside 'ships)
    (put riverside 'island)
    (put riverside 'river)
    (put riverside 'ferry)
    (put riverside 'ferryman)
    (put riverside 'warehouse)
    (put riverside 'industry-avenue)
    (set-current-location 'riverside))

  (let ((desk (def-prop 'desk "the desk")))
    (obj-pat '(seq (opt the) (or desk bureau)) 'desk)
    (act desk 'look (msg "This desk was once used by someone to keep track of the wood that was shipped in and out of here. At "
			 "the moment business seems to be nonexistent. There is something about the dust on the desk that "
			 "catches your attention."))
    (act desk 'climb (msg "You climb on the desk and stand there for a bit, but get bored with that game."))
    (act desk 'look-close look-close-at-desk))

  (let ((cabinet (def-prop 'cabinet "the cabinet")))
    (obj-pat '(seq (opt the) cabinet) 'cabinet)
    (act cabinet 'look look-in-cabinet)
    (act cabinet 'look-close look-in-cabinet))

  (let ((dust (def-prop 'dust "the dust")))
    (obj-pat '(seq (opt the) dust) 'dust)
    (act dust 'look (msg "Judging by the handprints in the dust someone has recently grabbed the desk. Patterns in the dirt on "
			 "the floor suggest that this someone moved it a little away from the wall."))
    (act dust 'get (msg "You pick up some dust, but all that accomplishes is getting your hands dirty.")))

  (let ((can (def-mov 'can "the empty can")))
    (obj-pat '(seq (opt the) (opt empty) can) 'can)
    (act can 'look (msg "The can smells like lamp oil. How very suspicious. A bramble leaf is sticking to the bottom.")))

  (let ((newspaper (def-mov 'newspaper "the newspaper")))
    (obj-pat '(seq (opt the) (opt old) newspaper) 'newspaper)
    (act newspaper 'look (msg "An old yellowed newspaper"))
    (act newspaper 'read (msg "This is a local paper dated the 4th of November 1864. Only five months ago. The front page has a "
			      "story about a number of houses burning down in the east of Dunwich. It suggests the fire was "
			      "started intentionally.")))

  (let ((ladder (def-mov 'ladder "the ladder")))
    (obj-pat '(seq (opt the) ladder) 'ladder)
    (act ladder 'look (msg "This is a light wooden ladder about two meters high.")))

  (let ((cobwebs (def-prop 'cobwebs "the cobwebs")))
    (obj-pat '(seq (opt the) (opt (or cob spider)) (or web webs cobwebs)) 'cobwebs)
    (act cobwebs 'look (msg "These are old cobwebs. The kind with loads of dust on them that no self-respecting spider would "
			    "want to inhabit anymore."))
    (act cobwebs 'get (msg "They are hanging too high. and you do not really think you want them anyway.")))

  (let ((beams (def-prop 'beams "the beams")))
    (obj-pat '(seq (opt the) (or beam beams)) 'beams)
    (act beams 'look (msg "They are just old wooden beams.")))

  (let ((sawdust (def-prop 'sawdust "the sawdust")))
    (obj-pat '(seq (opt the) sawdust) 'sawdust)
    (act sawdust 'look (msg "A thick layer of old sawdust covers the ground. At some point this warehouse must have been "
			    "heavily used."))
    (act sawdust 'get (msg "The sawdust is completely useless. Don't bother.")))

  (let ((warehouse (def-loc 'warehouse "the warehouse")))
    (obj-pat '(seq (opt the) (or warehouse warehouses)) 'warehouse)
    (act warehouse 'look-loc (msg "It is dark and dusty in the warehouse. There is nothing stored here and it looks like nothing "
				  "has been for quite a while. Huge cobwebs hang from the beams overhead and the sawdust on the "
				  "ground does not look recent. There are a desk and a cabinet in the far corner."))
    (act warehouse 'look (msg "The open warehouse seems to belong to the Dunwich lumber company. It looks empty."))
    (put warehouse 'desk)
    (put warehouse 'cabinet)
    (put warehouse 'dust)
    (put warehouse 'cobwebs)
    (put warehouse 'beams)
    (put warehouse 'ladder)
    (put warehouse 'sawdust)
    (put warehouse 'riverside))

  (let ((fence (def-prop 'fence "the wooden fence")))
    (obj-pat '(seq (opt the) (opt high) (opt wooden) fence) 'fence)
    (act fence 'look (msg "The fence is of better craftsmanship than most of the constructions around here. It is about two and a "
			  "half meter high."))
    (act fence 'climb (msg "You jump to the rim of the fence and pull yourself up. Looking over it, you see a pair of mean-looking "
			   "dogs. They look at you and start drooling expectantly. You quickly let yourself drop to the ground "
			   "on the street side again.")))

  (let ((houses (def-prop 'houses "the houses")))
    (obj-pat '(seq (opt the) (or house houses)) 'houses)
    (act houses 'look (msg "Some sad little worker houses. They are locked and no one seems to be home.")))

  (let ((officer (def-npc 'officer "the police officer")))
    (obj-pat '(seq (opt the) (opt police) (or officer man policeman cop)) 'officer)
    (act officer 'look (msg "A serious-looking police officer. Noticing you looking at him he says 'Hello there. I you do not "
			    "mind I have a question for you.'"))
    (act officer 'talk talk-to-officer)
    (act officer 'accept officer-accept)
    (add-turn-event move-officer 'move-officer))

  (let ((industry-avenue (def-loc 'industry-avenue "Industry Avenue")))
    (obj-pat '(seq (opt industry) (or avenue street road)) 'industry-avenue)
    (act industry-avenue 'look-loc (msg "This is Industry Avenue, a road of mossy cobblestones. It runs from the docks to the "
					"Gray Square in the north of the area. On one side of the street is a high wooden fence "
					"that runs around the cloth factory, on the other side are some houses and a lumber mill."))
    (act industry-avenue 'look (msg "The avenue goes north-south through South End. It is the only proper street there is in "
				    "this dump."))
    (act industry-avenue 'enter (explore-function 'gray-square 'lumber-mill))
    (put industry-avenue 'fence)
    (put industry-avenue 'cloth-factory)
    (put industry-avenue 'officer)
    (put industry-avenue 'riverside)
    (put industry-avenue 'gray-square)
    (put industry-avenue 'houses)
    (put industry-avenue 'lumber-mill))

  (let ((machinery (def-prop 'machinery "the machinery")))
    (obj-pat '(seq (opt the) (or machinery saw saws)) 'machinery)
    (act machinery 'look (msg "A large system of cogs and rods intended to power a few big metal saws.")))

  (let ((lumber-mill (def-loc 'lumber-mill "the lumber mill")))
    (obj-pat '(seq (opt the) (opt (or wood lumber)) mill) 'lumber-mill)
    (act lumber-mill 'look-loc (msg "This mostly wooden structure contains elaborate machinery to operate some saws. It looks "
				    "like it has fallen into disrepair. There is a little attic above you, but the entrance "
				    "to it is just a little too high to reach."))
    (act lumber-mill 'look (msg "The mill looks unused. The locks on its doors have long ago been broken open."))
    (act lumber-mill 'enter (explore-function 'attic))
    (put lumber-mill 'machinery)
    (put lumber-mill 'attic)
    (put lumber-mill 'industry-avenue))

  (let ((diary (def-mov 'diary "the diary")))
    (obj-pat '(seq (opt the) (opt alchemists) diary) 'diary)
    (act diary 'look (msg "This slim book has the word 'diary' and a symbol of a rose and a cross on its cover. Inside it are "
			  "page after page of handwritten strange symbols."))
    (act diary 'read try-to-read-diary))

  (let ((attic (def-loc 'attic "the attic")))
    (obj-pat '(seq (opt the) (opt lumber) (opt mill) (or attic loft)) 'attic)
    (act attic 'look-loc (msg "You are in the attic of the lumber mill. It is a ramshackle wooden construction - you can see "
			      "through the cracks in the floor. You'd rather not to stay here for too long."))
    (act attic 'look (msg ""))
    (act attic 'climb try-to-climb-attic)
    (act attic 'enter try-to-enter-attic)
    (put attic 'diary)
    (put attic 'lumber-mill))

  (let ((railroad (def-prop 'railroad "the railroad")))
    (obj-pat '(seq (opt the) (opt big) (opt elevated) (or railroad (seq rail road))) 'railroad)
    (act railroad 'look (msg "On an elevation of at least five meters lies the railroad. Trains pass noisily now and then."))
    (act railroad 'climb (msg "The railroad is broad and full of speeding trains. If you value your life you stay off it.")))

  (let ((gray-square (def-loc 'gray-square "Gray Square")))
    (obj-pat '(seq (opt gray) (or square place)) 'gray-square)
    (act gray-square 'look-loc (msg "Gray Square, the center of the residential part of this neighborhood. It was cut in half "
				    "a few years ago by a the construction of a big elevated railroad. Little houses surround you, "
				    "and on your left are a bakery and an entrance to the yard of the cloth factory. Industry "
				    "Avenue leads back to the river and a trail follows the railroad to the right."))
    (act gray-square 'enter (explore-function 'trail 'bakery 'yard))
    (act gray-square 'look (msg "It is a more or less square place covered with uneven cobblestones."))
    (put gray-square 'industry-avenue)
    (put gray-square 'railroad)
    (put gray-square 'houses)
    (put gray-square 'bakery)
    (put gray-square 'yard)
    (put gray-square 'trail))

  (let ((bread (def-prop 'bread "the bread")))
    (obj-pat '(seq (opt the) bread) 'bread)
    (act bread 'look (msg "The bread looks crude and cheap."))
    (act bread 'get (msg "You'd have to buy some.")))

  (let ((baker (def-npc 'baker "the baker")))
    (obj-pat '(seq (opt the) baker) 'baker)
    (act baker 'look (msg "For a guy who sells food, this baker sure is skinny. He is wearing clothes that probably once were "
			  "white, now they are more of an uneven gray color."))
    (act baker 'talk talk-to-baker))

  (let ((bakery (def-loc 'bakery "the bakery")))
    (obj-pat '(seq (opt the) (or bakery store)) 'bakery)
    (act bakery 'look-loc (msg "This is a sad little bakery. The bread looks of a horrible quality."))
    (act bakery 'look (msg "The bakery does not look like it has anything fancy. It is open though."))
    (put bakery 'bread)
    (put bakery 'baker)
    (put bakery 'gray-square))

  (let ((bum (def-npc 'bum "the bum")))
    (obj-pat '(or (seq (opt the) (opt old) bum) (seq (opt (seq jean paul)) (opt van) ricken)) 'bum)
    (act bum 'look (msg "You see an old bum sitting on the ground here. He looks at you expectingly, obviously hoping you'll "
			"start a conversation. There's an almost empty bottle of wine next to him. You notice a bandage around "
			"the little finger on his left hand."))
    (act bum 'talk talk-to-bum))

  (let ((wine (def-prop 'wine "the bottle of wine")))
    (obj-pat '(seq (opt the) (opt (seq bottle of)) (or wine bottle)) 'wine)
    (act wine 'look (msg "An almost empty bottle of cheap wine."))
    (act wine 'get (msg "The bum is not interested in sharing his wine with you.")))

  (let ((bandage (def-prop 'bandage "the dirty bandage")))
    (obj-pat '(seq (opt the) (or (seq (opt left) (opt little) finger) (seq (opt dirty) bandage))) 'bandage)
    (act bandage 'look (msg "A dirty bandage is wrapped around the bum's left little finger."))
    (act bandage 'get (msg "Leave the bandage alone, maniac.")))

  (let ((trail (def-loc 'trail "the trail")))
    (obj-pat '(seq (opt the) (or path trail)) 'trail)
    (act trail 'look-loc (msg "You are on a trail following the railroad. Gray Square lies on one end, the railroad bridge on "
			      "the other. There seems to be a little muddy area under the bridge."))
    (act trail 'look (msg "a muddy trail along the railroad"))
    (act trail 'enter (explore-function 'bridge))
    (put trail 'railroad)
    (put trail 'bum)
    (put trail 'bandage)
    (put trail 'wine)
    (put trail 'gray-square)
    (put trail 'bridge))

  (let ((building (def-prop 'building "the building")))
    (obj-pat '(seq (opt the) (opt huge) (opt brick) building) 'building)
    (act building 'look (msg "This building blocks the way to the other side of the railroad. It is probably some kind "
			     "of factory.")))

  (let ((brambles (def-prop 'brambles "the bramble bushes")))
    (obj-pat '(seq (opt the) (or brambles bushes (seq bramble bushes))) 'brambles)
    (act brambles 'look (msg "A dense mess of bramble bushes."))
    (act brambles 'get (msg "Nah, the thorns would hurt you."))
    (act brambles 'look-close look-close-at-brambles))

  (let ((matches (def-mov 'matches "the set of matches")))
    (obj-pat '(seq (opt the) (opt (seq set of)) (or matches match)) 'matches)
    (act matches 'look (msg "This set of matches has been out here for only a few days. They have gotten wet and are unusable, "
			    "but not very weathered yet.")))

  (let ((gloves (def-mov 'gloves "the pair of gloves")))
    (obj-pat '(seq (opt the) (opt (seq pair of)) gloves) 'gloves)
    (act gloves 'look (msg "A pair of black woolen gloves. The left glove is somewhat burned at the little finger.")))

  (let ((bridge (def-loc 'bridge "the railroad bridge")))
    (obj-pat '(seq (opt the) (or (seq muddy area) (seq (opt (or railroad (seq rail road))) bridge))) 'bridge)
    (act bridge 'look-loc (msg "You are under the railroad bridge that spans the Urth river. It is quiet here. Bramble bushes "
			       "grow against the bridge's pillars. On one side a trail leads back to the square, the other side "
			       "is blocked by the back of a huge brick building. The water level is only a little below you, this "
			       "area is probably underwater in winter."))
    (act bridge 'look (msg "A high metal-and-wood bridge to get trains to the other side of the river."))
    (act bridge 'climb (msg "The pillars of the bridge are too smooth to climb."))
    (put bridge 'river)
    (put bridge 'island)
    (put bridge 'railroad)
    (put bridge 'brambles)
    (put bridge 'building)
    (put bridge 'trail))

  (let ((guard (def-npc 'guard "a hired guard")))
    (obj-pat '(seq (opt the) (opt hired) guard) 'guard)
    (act guard 'look (msg "This is a big (very big) man. He is leaning on a rusty gun and staring at you all the time."))
    (act guard 'talk talk-to-guard)
    (act guard 'accept guard-accept))

  (let ((yard (def-loc 'yard "The yard of the cloth factory")))
    (obj-pat '(seq (opt the) (opt (seq (opt cloth) factory)) yard (opt (seq of (opt the) (opt cloth) factory))) 'yard)
    (act yard 'look-loc (msg "You are in the yard of the cloth factory. The building itself stands to the east and the Gray "
			     "Square lies to the north. People seem to be working inside, you hear voices and the noise of "
			     "weaving machines."))
    (act yard 'enter (explore-function 'cloth-factory))
    (put yard 'guard)
    (put yard 'cloth-factory)
    (put yard 'gray-square))

  (let ((machines (def-prop 'machines "the weaving machines")))
    (obj-pat '(seq (opt the) (opt weaving) (or machine machines)) 'machines)
    (act machines 'look (msg "These machines, powered by a central steam engine, weave wool into cloth.")))

  (let ((engine (def-prop 'engine "the steam engine")))
    (obj-pat '(seq (opt the) (opt steam) engine) 'engine)
    (act engine 'look (msg "A big steam engine to power the weaving machines.")))

  (let ((worker (def-npc 'worker "workers")))
    (obj-pat '(seq (opt the) (or worker workers)) 'worker)
    (act worker 'look (msg "These people look as if they work a lot for a lousy pay."))
    (act worker 'talk talk-to-worker))

  (let ((overseer (def-npc 'overseer "overseer")))
    (obj-pat '(seq (opt the) overseer) 'overseer)
    (act overseer 'look (msg "The overseer is a skinny little man. Yet somehow all the workers seem to be very nervous around him."))
    (act overseer 'talk talk-to-overseer))

  (let ((needle (def-mov 'needle "the needle")))
    (obj-pat '(seq (opt the) (or needle pin)) 'needle)
    (act needle 'look (msg "This is a small, very thin needle.")))

  (let ((cloth-factory (def-loc 'cloth-factory "the cloth factory")))
    (obj-pat '(seq (opt the) (opt (or cloth wool)) factory) 'cloth-factory)
    (act cloth-factory 'look-loc (msg "Inside the cloth factory. A big space with weaving machines lined up neatly. Almost every "
				      "machine has a worker at it and there is a lot of activity hauling wool and cloth around."))
    (act cloth-factory 'look (msg "The factory is a big building of red brick. It seems newer than most of the things here."))
    (act cloth-factory 'enter try-to-enter-factory)
    (put cloth-factory 'machines)
    (put cloth-factory 'engine)
    (put cloth-factory 'needle)
    (put cloth-factory 'worker)
    (put cloth-factory 'yard))
  )

(define (init-grammar)
  (define action define-action-pattern)
  (action '(or exit quit) quit-action)
  (action '(seq (or climb (seq jump (or on to for at)))(opt (or on over)) (object)) climb-action)
  (action '(seq open (object)) open-action)
  (action '(seq put (object) in (object)) put-in-action)
  (action '(or (seq (or fuck damn) (opt (or it you))) shit (seq get lost)) 
	  (message-function "Now now, no need to get angry. This can't be that hard."))
  (action '(seq (or kill shoot attack kick murder) (object)) 
	  (lambda (x) (print-message "I know it is kind of weird, but in this computer game killing things is not the solution.")))
  (action '(or north east south west) (message-function "Moving in this game works differently than in most text adventures. Type "
							"'help' to read more."))
  (action '(seq eat (object)) (lambda (x) (print-message "Eating " (find-name x) " is a bad idea.")))
  (action '(seq load (any)) load-action)
  (action '(seq save (any)) save-action)
  (action '(seq (or shout yell scream) (any)) (lambda (x) (print-message "Shhhhh. get a hold of yourself.")))
  (action '(seq (or look l check see) (or behind under between in) (object)) look-close-action)
  (action '(or help hint hints) help-action)
  (action '(or leave (seq (or run go) away)) leave-action)
  (action '(seq read (opt (or from in)) (object)) read-action)
  (action '(seq read entry (any)) read-diary-entry)
  (action '(seq read (any entry)) read-diary-entry)
  (action '(seq give (object) to (object)) give-action)
  (action '(seq say (or (seq quote (any quote)) (any))) say-action)
  (action '(seq ask (object) (or about for) (or (seq quote (any quote)) (any))) (lambda (x y) (talk-action y x)))
  (action '(seq (or say tell mention) (any to) (object)) talk-action)
  (action '(seq (or say tell ask mention) quote (any quote) to (object)) talk-action)
  (action '(or i inv inventory) inventory-action)
  (action '(seq (or drop lose dump (seq get rid of)) (object)) drop-action)
  (action '(seq (or get take grab) (object)) take-action)
  (action '(seq (or (seq (or go move m walk run) (opt to)) goto enter) (object)) move-action)
  (action '(seq (or l look x examine) (opt (seq (opt at) (object)))) look-action)
  )

(define (init)
  (reset-data)
  (init-grammar)
  (init-world))

; Init and start the game.
(define (main)
  (init)
  (game))

(main)
