;; Code for reading a string from the standard input and transforming
;; it into a list of lowercase symbols. Any non-alphanumeric character
;; is interpreted as whitespace. Use the (read-input-line) function.
;; (print-message . args) converts its arguments into a string and
;; prints it to the standard output, breaking lines between words.

; Line width for output.
(define line-width 79)

; Read a line into a raw string.
(define (read-raw-line)
  (let* ((buffer-size 1024) (buffer (make-string buffer-size #\newline)))
    (define (loop pos)
      (string-set! buffer pos (read-char))
      (cond ((eqv? (string-ref buffer pos) #\newline) buffer)
	    ((= pos (- buffer-size 2)) buffer) ; end of buffer
	    (else (loop (+ 1 pos)))))
    (loop 0)))

; Transform a character to lower case (if applicable).
(define (lowercase char)
  (if (char-upper-case? char)
      (char-downcase char)
      char))

; Transform a raw string into a list of symbols. This thing contains
; quite a bit of imperative programming. I guess I could have solved
; it with a more functional approach, but now that it works I'm not
; touching it anymore.
(define (split-line line)
  (let ((pos 0))
    (define (next-char) (set! pos (+ 1 pos)))
    (define (part-of-identifyer char) (or (char-alphabetic? char) (char-numeric? char)))
    (define (read-word) 
      (let ((current-char (string-ref line pos)))
	(cond ((part-of-identifyer current-char)
	       (next-char)
	       (cons (lowercase current-char) (read-word)))
	      (else '()))))
    (define (loop)
      (let ((current-char (string-ref line pos)))
	(cond ((eq? current-char #\newline) '())
	      ((member current-char '(#\" #\' #\`)) (next-char) (cons 'quote (loop)))
	      ((part-of-identifyer current-char) (cons (string->symbol (list->string (read-word))) (loop)))
	      (else (next-char) (loop)))))
    (loop)))

; Read a raw string and split it into a list of symbols.
(define (read-input-line)
  (split-line (read-raw-line)))

; Display a list as a string, taking line width into account and
; trying to break lines on spaces or hyphens.
(define (print-message . message)
  (print-string (build-string message)))

; Turn a list of strings, characters, symbols and numbers into a
; single string.
(define (build-string message)
  (cond ((null? message) "")
	((number? (car message)) (string-append (number->string (car message)) (build-string (cdr message))))
	((symbol? (car message)) (string-append (symbol->string (car message)) (build-string (cdr message))))
	((char? (car message)) (string-append (list->string (list (car message))) (build-string (cdr message))))
	((string? (car message)) (string-append (car message) (build-string (cdr message))))
	(else (print-message "error: wrong type being given to print-message"))))

; Print a string with a certain line width, to prevent words being cut
; in half at the end of the line.
(define (print-string string)
  (define (find-line-end string)
    (let ((last-break 0))
      (define (loop pos)
	(cond ((>= pos (string-length string)) (string-length string))
	      ((>= pos line-width) (if (= 0 last-break) pos last-break))
	      (else (if (member (string-ref string pos) '(#\- #\space))
			(set! last-break (+ 1 pos)))
		    (loop (+ 1 pos)))))
      (loop 0)))
  (let ((line-end (find-line-end string)))
    (display (substring string 0 line-end))
    (newline)
    (if (< line-end (string-length string))
	(print-string (substring string line-end (string-length string))))))

(define (capitalize string)
  (if (char-lower-case? (string-ref string 0))
      (let ((new-string (string-copy string)))
	(string-set! new-string 0 (char-upcase (string-ref string 0)))
	new-string)
      string))
