;; Nested tables, entity names in the top table, properties of those
;; entities below it.

; Create an empty table. Just gives a list of one element, the value
; of the element is not even very important. The other table functions
; expect every table (and subtable!) to start with an identifier, so
; this is an identifier with no elements after it.
(define (create-table)
  (list 'table))

; Insert element into a table. The first argument is a list of
; identifiers denoting the position in the table. This can be
; arbitrarily long, and none of the identifiers have to exist (they
; will be created). If an element with the same identifiers already
; exists in the table, the new value replaces the old one.
(define (insert-in-table type data table)
  (define (insert-in-headed-list value list)
    (set-cdr! list (cons value (cdr list))))
  (let ((found (assoc (car type) (cdr table))))
    (if (not found)
	(begin (insert-in-headed-list (list (car type)) table)
	       (set! found (cadr table))))
    (if (null? (cdr type))
	(set-cdr! found data)
	(insert-in-table (cdr type) data found))))

; Find an element in a table. The type argument is of the same kind as
; with the (insert-in-table) function. Returns #f if no element with
; those identifiers exists.
(define (lookup-in-table type table)
  (let ((found (assoc (car type) (cdr table))))
    (cond ((not found) #f)
	  ((null? (cdr type)) found)
	  (else (lookup-in-table (cdr type) found)))))

; Insert the content of a table into another table.
(define (merge-table addition base-table)
  (define (loop-elements list root)
    (cond ((not (null? list))
	   (handle-element (car list) root)
	   (loop-elements (cdr list) root))))
  (define (handle-element element root)
    (cond ((not (list? element)) (insert-in-table (append root (list (car element))) (cdr element) base-table))
	  (else (loop-elements (cdr element) (append root (list (car element)))))))
  (loop-elements (cdr addition) '()))

; Remove all (identifier . value) pairs that do not satify a predicate
; from a table.
(define (filter-table predicate table)
  (cond ((null? table) '())
	((list? table) (let ((filtered (filter-table predicate (car table))))
			 (if (null? filtered)
			     (filter-table predicate (cdr table))
			     (cons filtered (filter-table predicate (cdr table))))))
	((pair? table) (if (predicate table)
			   table
			   '()))
	(else table)))
