;; The parse function. Used to match sentences to actions and to
;; extract the significant parts from those sentences.

;; The patterns that this parser recognizes are combinations of the
;; following:
;; <symbol>: A symbol matches with itself.
;; (seq <a> <b> ....): Matches to a sequence of patterns.
;; (opt <pattern>): Matches to either <pattern> or nothing.
;; (or <a> <b> ...): Matches to one from a number of patterns.
;; (any <pattern>): Matches to anything until it encounters something
;;    that matches <pattern>. Stores the matched words.
;; (object): Matches to any of the object patterns (and stores that 
;;    object as an argument to the action).

;; PARSE OBJECT
;; A way to combine the remainer of the parsed sentence and the
;; significant parts (object idenifiers, (any) matches) found in it in
;; return values. Only needs a few operations (removing from the
;; remainder, consing to the found objects, extracting parts).

(define (create-parse-object input)
  (cons input '()))

(define (remain parse-object)
  (car parse-object))

(define (found parse-object)
  (cdr parse-object))

(define (clear-remain parse-object)
  (cons '() (found parse-object)))

(define (r-cdr parse-object)
  (cons (cdr (car parse-object)) (cdr parse-object)))

(define (f-cons new-found parse-object)
  (cons (car parse-object) (cons new-found (cdr parse-object))))

;; PARSE FUNCTION
;; This one is called from outside.

; Arguments are the input (list of symbols) and a set of patterns.
; The set is a list of pairs, every pair indicates a pattern. The car
; of the pair is the pattern, and the cdr the function it is
; associated with.
(define (parse input patterns)
  (cond ((null? patterns) #f)
	(else (let ((this-match (match-pattern (create-parse-object input) (caar patterns))))
		(if (and this-match (null? (remain this-match)))
		    (list (cdar patterns) (reverse (found this-match)))
		    (parse input (cdr patterns)))))))

;; MATCH FUNCTIONS
;; Called by parse to do the actual pattern matching.

; Match the start of a sentence with a pattern. If the pattern
; matches this returns the remaining part of the sentence,
; otherwise it returns false.
(define (match-pattern input pattern)
  (cond ((null? pattern) #f)
	((or (symbol? pattern) (number? pattern)) (match-symbol input pattern))
	(else (case (car pattern)
		((or) (match-or input (cdr pattern)))
		((seq) (match-seq input (cdr pattern)))
		((opt) (match-opt input (cdr pattern)))
		((any) (match-any input (cdr pattern)))
		((object) (match-object input))
		(else (display "unknown pattern (match-pattern)") (newline))))))

; See if the input starts with a given symbol.
(define (match-symbol input pattern)
  (cond ((null? (remain input)) #f)
	((eqv? (car (remain input)) pattern) (r-cdr input))
	(else #f)))

; Allow the input to start with one of a list of patterns.
(define (match-or input pattern)
  (cond ((null? pattern) #f)
	((match-pattern input (car pattern)))
	(else (match-or input (cdr pattern)))))

; Allow a sequence of patterns.
(define (match-seq input pattern)
  (if (null? pattern)
      input
      (let ((match (match-pattern input (car pattern))))
	(if match (match-seq match (cdr pattern)) #f))))

; Match with the pattern but no problem if it does not match.
(define (match-opt input pattern)
  (let ((match (match-pattern input (car pattern))))
    (if match match input)))

; Match anything (other than '()), until pattern is found. The rather
; clumsy form of requiring an ending pattern is needed to decide where
; the end of the match is. If none is given, this will match the rest
; of the sentence.
(define (match-any input pattern)
  (cond ((null? (remain input)) #f)
	((null? pattern) (f-cons (remain input) (clear-remain input)))
	(else
	 (let ((accum-any (collector)))
	   (define (match-pattern-any input pattern)
	     (cond ((null? (remain input)) #f)
		   (else (accum-any (car (remain input)))
			 (cond ((match-pattern (r-cdr input) pattern))
			       (else (match-pattern-any (r-cdr input) pattern))))))
	   (let ((retval (match-pattern-any input (car pattern))))
	     (if retval
		 (f-cons (accum-any) retval)
		 #f))))))

; Try to match an object and store it if found. Note that this thing
; relies on the (object-patterns) function, thus kind of ruining the
; independance of this parse module.
(define (match-object input)
  (define (match-pattern-object input patterns)
    (if (null? patterns)
	#f
	(let ((match (match-pattern input (caar patterns))))
	  (if match
	      (f-cons (cdar patterns) match)
	      (match-pattern-object input (cdr patterns))))))
  (match-pattern-object input (object-patterns)))
